/* ez-blob.h: Serialising description and serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>

#include "ez-blob.h"

/*
 This implements a basic 'high performance' serialisation based on the
 descriptors.

 Fields are encoded in the same order as the descriptor with no
 labelling.  Therefore the same descriptor must be used on both ends.
 Fields may be at arbitrary alignments.

 primitives
  encoded in machine order at native size

 strings
  4-byte length followed by characters.  No nul termination.  A length
  of ~0 indicates a null string pointer.

 arrays
  4-byte length followed by data.  ** currently null pointer arrays with zero length are converted to a zero-length array.

 struct
  4-byte length followed by blob-encoded data.

 pointer
  4-byte length followed by blob-encoded data.  A length of ~0
  indicates a null pointer.

*/

static const int primitive_size[EZ_BLOB_FLOAT64+1] = {
	1, 2, 4, 8,
	4, 8
};

size_t ez_blob_size(const ez_blob_desc *d, const void *p) {
	size_t size = 0;

	for (;(d->bd_type != EZ_BLOB_END);d++) {
		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			size += primitive_size[d->bd_type];
		} else {
			const void *v = p + d->bd_offset;
			switch (d->bd_type) {
			case EZ_BLOB_STRING: {
				char *s = ((char **)v)[0];
				size += s ? 4 + strlen(s) : 4;
				break; }
			case EZ_BLOB_ARRAY:
				size += 4 + ((const struct ez_blob_array *)v)->size;
				break;
			case EZ_BLOB_STRUCT: {
				const void *sp = v;
				size += 4 + ez_blob_size(d->bd_table, sp);
				break; }
			case EZ_BLOB_POINTER: {
				const void *s = ((const void * const *)v)[0];
				size += s ? 4 + ez_blob_size(d->bd_table, s) : 4;
				break; }
			default:
				break;
			}
		}
	}

	return size;
}

void ez_blob_free_raw(const ez_blob_desc *d, void *p) {
	for (;(d->bd_type != EZ_BLOB_END);d++) {
		void *v = p + d->bd_offset;
		
		switch (d->bd_type) {
		case EZ_BLOB_STRING:
			free(((char **)v)[0]);
			break;
		case EZ_BLOB_ARRAY:
			free(((struct ez_blob_array *)v)->data);
			break;
		case EZ_BLOB_STRUCT:
			ez_blob_free_raw(d->bd_table, v);
			break;
		case EZ_BLOB_POINTER:
			ez_blob_free(d->bd_table, ((void **)v)[0]);
			break;
		default:
			break;
		}
	}
}

void ez_blob_free(const ez_blob_desc *d, void *p) {
	if (p) {
		ez_blob_free_raw(d, p);
		free(p);
	}
}

void *ez_blob_decode_raw(const ez_blob_desc *d, const char *b, size_t size, void *p) {
	const char *be = b + size;

	memset(p, 0, d->bd_offset);
	
	for (;(d->bd_type != EZ_BLOB_END);d++) {
		void *v = (p + d->bd_offset);

		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			int psize = primitive_size[d->bd_type];

			if (b + psize > be)
				goto fail;
			memcpy(v, b, psize);
			b += psize;
		} else if (d->bd_type <= EZ_BLOB_LAST) {
			uint32_t ss;

			if (b + 4 > be)
				goto fail;
			memcpy(&ss, b, 4);
			b += 4;

			switch (d->bd_type) {
			case EZ_BLOB_STRING: {
				char *s = NULL;

				if (ss != ~0) {
					if (b + ss > be)
						goto fail;
					if ((s = malloc((size_t)ss + 1)) == NULL)
						goto fail;
					memcpy(s, b, ss);
					s[ss] = 0;
					b += ss;
				}
				((char **)v)[0] = s;
				break; }
			case EZ_BLOB_ARRAY: {
				struct ez_blob_array *a = v;
				
				if (b + ss > be)
					goto fail;
				a->size = ss;
				if ((a->data = malloc(ss)) == NULL)
					goto fail;
				memcpy(a->data, b, ss);
				b += ss;
				break; }
			case EZ_BLOB_STRUCT:
				if (b + ss > be)
					goto fail;
				
				if (ez_blob_decode_raw(d->bd_table, b, ss, v) == NULL)
					goto fail;
				b += ss;
				break;
			case EZ_BLOB_POINTER: {
				void *s = NULL;

				if (ss != ~0) {
					if (b + ss > be)
						goto fail;
					if ((s = calloc(d->bd_table->bd_offset, 1)) == NULL)
						goto fail;
					if (ez_blob_decode_raw(d->bd_table, b, ss, s) == NULL) {
						free(s);
						goto fail;
					}
					b += ss;
				}
				((void **)v)[0] = s;
				break; }
			default:
				break;
			}
		}
	}
	if (b != be)
		goto fail;

	return p;
 fail:
	ez_blob_free_raw(d, p);
	return NULL;
}

void *ez_blob_decode(const ez_blob_desc *d, const char *b, size_t size) {
	void *p = calloc(d->bd_offset, 1);

	if (ez_blob_decode_raw(d, b, size, p) == NULL) {
		free(p);
		return NULL;
	}
	
	return p;
}

void ez_blob_encode_raw(const ez_blob_desc *d, const void *p, char *b, size_t size) {
	char *be = b+size;
	for (;(d->bd_type != EZ_BLOB_END);d++) {
		const void *v = (p + d->bd_offset);

		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			int psize = primitive_size[d->bd_type];
			
			assert(b+psize <= be);
			memcpy(b, v, psize);
			b += psize;
		} else if (d->bd_type <= EZ_BLOB_LAST) {
			uint32_t ss;
			
			switch (d->bd_type) {
			case EZ_BLOB_STRING: {
				char *s = ((char **)v)[0];
				if (s) {
					ss = strlen(s);
					assert(b+4+ss <= be);
					memcpy(b, &ss, 4);
					memcpy(b+4, s, ss);
					b += 4 + ss;
				} else {
					ss = ~0;
					assert(b+4 <= be);
					memcpy(b, &ss, 4);
					b += 4;
				}
				break; }
			case EZ_BLOB_ARRAY: {
				const struct ez_blob_array *a = v;

				ss = a->size;
				assert(b+4+ss <= be);
				memcpy(b, &ss, 4);
				memcpy(b+4, a->data, ss);
				b += 4 + ss;
				break; }
			case EZ_BLOB_STRUCT:
				ss = ez_blob_size(d->bd_table, v);
			      
				assert(b+4+ss <= be);
				memcpy(b, &ss, 4);
				ez_blob_encode_raw(d->bd_table, v, b+4, ss);
				b += 4 + ss;
				break;
			case EZ_BLOB_POINTER: {
				const void *s = ((const void **)v)[0];

				if (s) {
					ss = ez_blob_size(d->bd_table, s);
					assert(b+4+ss <= be);
					memcpy(b, &ss, 4);
					ez_blob_encode_raw(d->bd_table, s, b+4, ss);
					b += 4 + ss;
				} else {
					ss = ~0;
					assert(b+4 <= be);
					memcpy(b, &ss, 4);
					b += 4;
				}
				break; }
			default:
				break;
			}
		}
	}
	assert(b == be);
}

/**
 * Encode blob to new buffer.
 *
 * @param p object to serialise.
 * @param sizep size of encoded blob return pointer.
 * @return a newly allocated blob of the correct size.
 */
char *ez_blob_encode(const ez_blob_desc *d, const void *p, size_t *sizep) {
	size_t size = ez_blob_size(d, p);
	char *b = malloc(size);

	ez_blob_encode_raw(d, p, b, size);

	*sizep = size;
	return b;
}

