/* ez-blob.h: Serialising description and serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_BLOB_H
#define _EZ_BLOB_H

#include <stddef.h>

/**
 * This is a C structure serialiser.
 *
 * It is table driven and can describe most common C data structures,
 * including nested and linked structures.
 *
 * The implemented serialisation mechanism only supports
 * forward-linked structures and will not perform well when
 * marshalling deeply nested structures but other serialisation
 * mechanisms are possible.
 */

/**
 *  An array of ez_blob_desc is passed to ez_blob_encode/decode.  The first
 *  entry must be of type EZ_BLOB_PK and it's 'offset' must be the size
 *  of the structure.  The final entry must be of type EZ_BLOB_END.
 * 
 *   As an example:
 * 
 * typedef struct dbdisk {
 * 	int id;
 * 	char *uuid;
 * 	char *label;
 * 	char *type;
 * } dbdisk;
 * 
 *   Has a table such as this, using the init macros.
 * 
 * static ez_blob_desc DISK_ENC[] = {
 * 	EZ_BLOB_START(dbdisk),
 * 	EZ_BLOB_STRING(dbdisk, 1, uuid),
 * 	EZ_BLOB_STRING(dbdisk, 2, label),
 * 	EZ_BLOB_STRING(dbdisk, 3, type),
 * 	EZ_BLOB_END(dbdisk)
 * };
 * 
 *   The 'id' field (of ez_blob_desc) is currently unused by the blob
 *   serialiser but could be implemented for data versioning or partial
 *   serialisation.
 * 
 */
typedef struct ez_blob_desc {
	unsigned short bd_type;		// type of field
	unsigned short bd_id;		// used to tag encoded entries (currently unused)
	unsigned int bd_offset;		// offset into structure, if type == EZ_BLOB_PK then this is sizeof(struct)
	union {
		const struct ez_blob_desc *table;// embedded structure or pointer
	} u;
#define bd_table u.table
} ez_blob_desc;

/**
 * Field type
 */
typedef enum ez_blob_desc_type {
	EZ_BLOB_INT8,
	EZ_BLOB_INT16,
	EZ_BLOB_INT32,
	EZ_BLOB_INT64,
	EZ_BLOB_FLOAT32,
	EZ_BLOB_FLOAT64,
	EZ_BLOB_STRING,			// normall c strring
	EZ_BLOB_ARRAY,			// embedded ez_blob_array for arbitrary bytes.
	EZ_BLOB_STRUCT,			// bd_table points to another descriptor list, object embedded
	EZ_BLOB_POINTER,		// bd_table points to another descriptor list, object pointer
	EZ_BLOB_LAST = EZ_BLOB_POINTER,
	EZ_BLOB_PK = 16,	 	// MUST always be first element in descriptor and occur only once
	EZ_BLOB_END
} ez_blob_desc_type;

typedef struct ez_blob_array {
	unsigned int size;
	void *data;
} ez_blob_array;

#define EZ_BLOB_START(s) { EZ_BLOB_PK, 0, sizeof(s) }
#define EZ_BLOB_INT8(s, id, f) { EZ_BLOB_INT8, id, offsetof(s, f) }
#define EZ_BLOB_INT16(s, id, f) { EZ_BLOB_INT16, id, offsetof(s, f) }
#define EZ_BLOB_INT32(s, id, f) { EZ_BLOB_INT32, id, offsetof(s, f) }
#define EZ_BLOB_INT64(s, id, f) { EZ_BLOB_INT64, id, offsetof(s, f) }
#define EZ_BLOB_FLOAT32(s, id, f) { EZ_BLOB_FLOAT32, id, offsetof(s, f) }
#define EZ_BLOB_FLOAT64(s, id, f) { EZ_BLOB_FLOAT64, id, offsetof(s, f) }
#define EZ_BLOB_STRING(s, id, f) { EZ_BLOB_STRING, id, offsetof(s, f) }
#define EZ_BLOB_ARRAY(s, id, f) { EZ_BLOB_ARRAY, id, offsetof(s, f) }
#define EZ_BLOB_STRUCT(s, id, f, other) {  EZ_BLOB_STRUCT, id, offsetof(s, f), .u.table = other }
#define EZ_BLOB_POINTER(s, id, f, other) { EZ_BLOB_POINTER, id, offsetof(s, f), .u.table = other }
#define EZ_BLOB_END(s) { EZ_BLOB_END }

/**
 * Calculate the serialised blob size for a struct.
 *
 * @param d descriptor table matching struct.
 * @param p struct.
 */
size_t ez_blob_size(const ez_blob_desc *d, const void *p);

/**
 * Marshal a struct to a blob.
 * 
 * @param d descriptor table matching the struct.
 * @param p structure.
 * @param sizep return pointer for blob size, must not be null.
 * @return the blob, it must be freed with free() when complete.
 */
char *ez_blob_encode(const ez_blob_desc *d, const void *p, size_t *sizep);

/**
 * Marshal a struct to a blob directly
 *
 * @param d descriptor table matching the struct.
 * @param p structure.
 * @param b target memory, must be as large as size
 * @param size size of target memory, must be exactly ez_blob_size().
 * @return the blob, it must be freed with free() when complete.
 */
void ez_blob_encode_raw(const ez_blob_desc *d, const void *p, char *b, size_t size);

/**
 * Unmarshall a blob into a struct.
 *
 * @param d descriptor table matching blob.
 * @param blob raw blob.
 * @param size blob size.
 * @return the decoded struct.  May be freed using ez_blob_free().
 */
void *ez_blob_decode(const ez_blob_desc *d, const char *blob, size_t size);

/**
 * Decode to pre-allocated buffer.
 *
 * @param b blob data
 * @param size blob size
 * @return p is returned on success.  On failure NULL is returned and any internal pointers are freed.
 */
void *ez_blob_decode_raw(const ez_blob_desc *d, const char *b, size_t size, void *p);

/**
 * Free struct unmarshalled by ez_blob_decode().
 *
 * @param d descriptor table matching struct.
 * @param p decoded struct from ez_blob_decode().
 */
void ez_blob_free(const ez_blob_desc *d, void *p);

/**
 * Free the contents of a structure, but don't free the container.
 *
 * @todo should this zero the structure too?
 * @param p structure to free.  p will not be freed.
 */
void ez_blob_free_raw(const ez_blob_desc *d, void *p);

#endif
