
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <assert.h>

#include "ez-blob.h"

struct simple {
	char b;
	short s;
	int a;
	float c;
	double d;
};

static const ez_blob_desc simple_DESC[] = {
	EZ_BLOB_START(struct simple),
	EZ_BLOB_INT8(struct simple, 1, b),
	EZ_BLOB_INT16(struct simple, 2, s),
	EZ_BLOB_INT32(struct simple, 3, a),
	EZ_BLOB_FLOAT32(struct simple, 4, c),
	EZ_BLOB_FLOAT64(struct simple, 5, d),
	EZ_BLOB_END(struct simple)
};

struct arrays {
	uint32_t count;
	char *string;
	ez_blob_array array;
};

static const ez_blob_desc array_DESC[] = {
	EZ_BLOB_START(struct arrays),
	EZ_BLOB_STRING(struct arrays, 1, string),
	EZ_BLOB_ARRAY(struct arrays, 2, array),
	EZ_BLOB_END(struct arrays)
};

struct embedded {
	struct simple s;
	struct arrays a;
};

static const ez_blob_desc embed_DESC[] = {
	EZ_BLOB_START(struct embedded),
	EZ_BLOB_STRUCT(struct embedded, 1, s, simple_DESC),
	EZ_BLOB_STRUCT(struct embedded, 2, a, array_DESC),
	EZ_BLOB_END(struct embedded)
};

struct tree {
	struct tree *left;
	struct tree *right;

	char *s;
};

static const ez_blob_desc tree_DESC[] = {
	EZ_BLOB_START(struct tree),
	EZ_BLOB_STRING(struct tree, 1, s),
	EZ_BLOB_POINTER(struct tree, 2, left, tree_DESC),
	EZ_BLOB_POINTER(struct tree, 3, right, tree_DESC),
	EZ_BLOB_END(struct tree)
};

static void dumphex(const char *data, size_t size, const char *prefix) {
	for (int i=0;i<size;i+=16) {
		fputs(prefix, stdout);
		printf(" %04x:", i);
		for (int j=0;j<16;j++) {
			if (i+j<size)
				printf(" %02x", data[i+j] & 0xff);
			else
				fputs("   ", stdout);
		}
		fputs("  ", stdout);
		for (int j=0;j<16;j++) {
			putchar(i+j<size && isprint(data[i+j]) ? data[i+j] : '.');
		}
		putchar('\n');
	}
}

static int blob_print(const ez_blob_desc *d, const void *a, int depth) {
	int e = 1;
	char x[depth+1];

	memset(x, ' ', depth);
	x[depth] = 0;
	
	for (;e && (d->bd_type != EZ_BLOB_END);d++) {
		const void *u = a + d->bd_offset;

		printf("%s [type=$%02x, offset=$%04x]", x, d->bd_type, d->bd_offset);
		
		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			printf(" .%d = %d\n", d->bd_id, *(uint8_t *)u);
			break;
		case EZ_BLOB_INT16:
			printf(" .%d = %d\n", d->bd_id, *(uint16_t *)u);
			break;
		case EZ_BLOB_INT32:
			printf(" .%d = %d\n", d->bd_id, *(uint32_t *)u);
			break;
		case EZ_BLOB_INT64:
			printf(" .%d = %ld\n", d->bd_id, *(uint64_t *)u);
			break;
		case EZ_BLOB_FLOAT32:
			printf(" .%d = %f\n", d->bd_id, *(float *)u);
			break;
		case EZ_BLOB_FLOAT64:
			printf(" .%d = %f\n", d->bd_id, *(double *)u);
			break;
		case EZ_BLOB_STRING:
			printf(" .%d %p = `%s'\n", d->bd_id, *((char **)u), *((char **)u));
			break;
		case EZ_BLOB_ARRAY: {
			const struct ez_blob_array *ua = u;
			
			printf(" .%d %p [%d] = {\n", d->bd_id, ua->data, ua->size);
			dumphex(ua->data, ua->size, x);
			printf("%s  }\n", x);
			break; }
		case EZ_BLOB_POINTER: {
			void *up = *((void **)u);
			if (up) {
				printf(" .%d %p = {\n", d->bd_id, up);
				blob_print(d->bd_table, up, depth+4);
				printf("%s  }\n", x);
			} else {
				printf(" .%d %p\n", d->bd_id, up);
			}
			break;
		}
		default:
			printf("\n");
		}
	}

	return e;
}

static int blob_equals(const ez_blob_desc *d, const void *a, const void *b) {
	int e = 1;
	for (;e && (d->bd_type != EZ_BLOB_END);d++) {
		const void *u = a + d->bd_offset;
		const void *v = b + d->bd_offset;

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			e &= (*(uint8_t *)u) == (*(uint8_t *)v);
			break;
		case EZ_BLOB_INT16:
			e &= (*(uint16_t *)u) == (*(uint16_t *)v);
			break;
		case EZ_BLOB_INT32:
			e &= (*(uint32_t *)u) == (*(uint32_t *)v);
			break;
		case EZ_BLOB_INT64:
			e &= (*(uint64_t *)u) == (*(uint64_t *)v);
			break;
		case EZ_BLOB_FLOAT32:
			e &= (*(float *)u) == (*(float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			e &= (*(double *)u) == (*(double *)v);
			break;
		case EZ_BLOB_STRING:
			e &= strcmp((*(char **)u), (*(char **)v)) == 0;
			break;
		case EZ_BLOB_ARRAY: {
			const struct ez_blob_array *ua = u;
			const struct ez_blob_array *va = v;

			e &= ua->size == va->size
				&& memcmp(ua->data, va->data, ua->size) == 0;
			break; }
		}
	}

	return e;
}

static void test_basics(const ez_blob_desc *d, void *src) {
	void *blob, *blob2;
	void *t;
	size_t size, size2;
	size_t dsize = d->bd_offset;
	char tmp[dsize + 64] __attribute__ ((aligned(64)));

	printf("source\n");
	blob_print(d, src, 4);
	
	blob = ez_blob_encode(d, src, &size);

	// check decode equals
	t = ez_blob_decode(d, blob, size);
	assert(t != NULL);
	assert(blob_equals(d, t, src));

	dumphex(blob, size, "serial: ");

	printf("decoded\n");
	blob_print(d, t, 4);

	// check encode-decoded equals source
	blob2 = ez_blob_encode(d, t, &size2);
	assert(size == size2);
	assert(memcmp(blob, blob2, size) == 0);
	free(blob2);
	
	ez_blob_free(d, t);

	// check raw decode stays within bounds of struct
	memset(tmp, 0xbe, sizeof(tmp));
	t = ez_blob_decode_raw(d, blob, size, &tmp[32]);
	//assert(t != NULL);
	//assert(blob_equals(d, tmp+32, src));
	dumphex(tmp, sizeof(tmp), "object: ");

	for (int i=0;i<32;i++) {
		assert((tmp[i] & 0xff) == 0xbe);
		assert((tmp[i + 32 + dsize] & 0xff) == 0xbe);
	}
	ez_blob_free_raw(d, tmp+32);
	free(blob);
}

static void test_simple(void) {
	struct simple src = {
		'z', 0xff0f, 1, 3.14159f, 5644941221133
	};
	const ez_blob_desc *d = simple_DESC;

	test_basics(d, &src);
}

static void test_arrays(void) {
	char data[7] = { 1, 2, 3, 4, 5, 6, 7 };
	struct arrays src = {
		32,
		"Bob was here.",
		.array.size = 7,
		.array.data = data
	};
	const ez_blob_desc *d = array_DESC;

	test_basics(d, &src);
}

static void test_tree(void) {
	struct tree src[6] = {
		{ &src[1], &src[3], "root" },
		{ NULL, &src[2], "left" },
		{ NULL, NULL, "left-right" },
		{ &src[4], &src[5], "right" },
		{ NULL, NULL, "right-left" },
		{ NULL, NULL, "right-right" },
	};
	const ez_blob_desc *d = tree_DESC;

	test_basics(d, &src);
}

int main(int argc, char **argv) {
	test_simple();
	test_arrays();
	test_tree();
}
